#!/usr/bin/env python3

import os
import sys
import logging
import json
import shutil
import glob
import errno
import argparse
import configparser
from pygears_tools.utils import (shell_source, custom_run, download_and_untar, list_pkg_deps,
                                 clone_git, install_deps, set_env, os_name, install_deps,
                                 pip_install)
from pygears_tools import default_cpp


def create_logger(pkg):
    logger = logging.getLogger(pkg["name"])
    logger.setLevel(logging.INFO)
    # create file handler which logs even debug messages

    # log_file = os.path.join(pkg["install_path"], "custom_cmd.log")
    # fh = logging.FileHandler(log_file, mode='w')
    # fh.setLevel(logging.INFO)
    # create console handler with a higher log level
    ch = logging.StreamHandler()
    ch.setLevel(logging.INFO)
    # create formatter and add it to the handlers
    formatter = logging.Formatter('%(asctime)s [%(name)-12s]: %(message)s', datefmt='%H:%M:%S')
    # fh.setFormatter(formatter)
    ch.setFormatter(formatter)
    # add the handlers to the logger
    # logger.addHandler(fh)
    logger.addHandler(ch)
    pkg["logger"] = logger


def os_install_cmd():
    name = os_name()
    if name == "ubuntu":
        return "sudo apt install -y"
    elif name == "opensuse":
        return "sudo zypper --non-interactive install"
    else:
        raise "Unsupported OS"


def filter_deps_by_os(pkgs):
    name = os_name()
    for pkg in pkgs:
        if ('deps' in pkg) and name in pkg['deps']:
            pkg['deps'] = pkg['deps'][name]


def copyanything(src, dst):
    try:
        shutil.copytree(src, dst)
    except OSError as exc:  # python >2.5
        if exc.errno == errno.ENOTDIR:
            shutil.copy(src, dst)
        else:
            raise


def get_pkg_classes(pkgs):
    classes = set()

    for pkg in pkgs:
        if 'class' in pkg:
            classes.update(pkg['class'])

    return classes


def print_pkg_deps_install(lines):
    for l in lines:
        print('{} {}'.format(os_install_cmd(), ' '.join(l)))


def expand_path(path):
    return os.path.abspath(os.path.expanduser(os.path.expandvars(path)))


def install(pkgs_fn, pkg_names, tools_path, home_path, do_install_deps, list_deps, dry_run):
    # print("Please enter sudo password:")
    # subprocess.run('sudo echo "Current install directory"; pwd', shell=True)

    cfg = {
        "home_path": expand_path(home_path) if home_path else None,
        "tools_path": expand_path(tools_path) if tools_path else None,
        "tools_install_path": expand_path(tools_path if tools_path else '~/.pygears/tools'),
        "install_script_path": expand_path(os.path.dirname(__file__)),
        "tools_sh_path": expand_path(os.path.join(tools_path, "tools.sh")) if tools_path else None,
        "pkgs_fn": expand_path(pkgs_fn),
        'dry_run': dry_run
    }

    if not list_deps:
        if cfg["tools_path"]:
            print(f'Installing to: {cfg["tools_path"]}')
        else:
            print(f'Installing system-wide')

    with open(cfg["pkgs_fn"]) as json_data:
        pkgs = json.load(json_data)

    if pkg_names:
        pkgs = [
            p for p in pkgs
            if (p['name'] in pkg_names or any(cls in pkg_names for cls in p['class']))
        ]

    filter_deps_by_os(pkgs)

    if list_deps:
        deps = list_pkg_deps(pkgs)
        print_pkg_deps_install(deps)
        return 0

    print(f'Storing installation files into: "{cfg["tools_install_path"]}". Creating folder.')
    if not cfg['dry_run']:
        if cfg['home_path']:
            os.makedirs(cfg["home_path"], exist_ok=True)

        os.makedirs(cfg["tools_install_path"], exist_ok=True)
        os.chdir(cfg["tools_install_path"])

    if cfg["tools_sh_path"] and not os.path.exists(cfg["tools_sh_path"]):
        with open(cfg["tools_sh_path"], "w") as text_file:
            print("#!/bin/bash", file=text_file)
            print("# Script for setting up the environment for all the tools", file=text_file)
            print("# Tools installed relative to: {}".format(cfg["tools_path"]), file=text_file)
            print("", file=text_file)

            print("# Setting new home directory:", file=text_file)

            if cfg['home_path']:
                print("export HOME={}".format(cfg["home_path"]), file=text_file)

    if not cfg['dry_run']:
        shell_source(cfg["tools_sh_path"])

    pyenv = False
    for i in range(len(pkgs)):
        if pkgs[i]['name'] == 'pyenv':
            if not cfg['tools_path']:
                del pkgs[i]
            else:
                pyenv = True

            break

    for pkg in pkgs:
        pkg.update(cfg)

        if pkg['tools_path']:
            pkg["path"] = expand_path(os.path.join(pkg["tools_install_path"], pkg["name"]))
        else:
            pkg["path"] = None

        pkg["install_path"] = expand_path(
            os.path.join(pkg["tools_install_path"], pkg["name"], "_install"))

        if not cfg['dry_run']:
            os.chdir(pkg["tools_install_path"])
            os.makedirs(pkg["name"], exist_ok=True)
            os.makedirs(pkg["install_path"], exist_ok=True)

        if "src_root_path" in pkg:
            pkg["src_root_path"] = pkg["src_root_path"].format(**pkg)

        create_logger(pkg)

    if do_install_deps:
        install_deps(pkgs)

    for pkg in pkgs:

        pkg["logger"].info("Installation started.")
        os.chdir(pkg["install_script_path"])

        if "copy" in pkg:
            pkg["logger"].info("Copying package files...")
            for cmd in pkg["copy"]:
                for fn in glob.glob(cmd[0].format(**pkg)):
                    pkg["logger"].info("Copying {} to {}".format(fn, cmd[1].format(**pkg)))

                    if not pkg['dry_run']:
                        copyanything(fn, os.path.join(cmd[1].format(**pkg), os.path.basename(fn)))

        if not pkg['dry_run']:
            os.chdir(pkg["install_path"])

        if "url" in pkg:
            download_and_untar(pkg)

        if "git" in pkg:
            clone_git(pkg)

        custom_run(pkg, "pre_custom_run")

        if pkg['path']:
            custom_run(pkg, "pre_custom_runl")

        if "flow" in pkg:
            pkg["logger"].info("Using {} flow.".format(pkg["flow"]))
            if pkg["flow"] == "default_cpp":
                default_cpp.flow(pkg)

        if "pip" in pkg:
            pip_install(pkg, pyenv)

        if not pkg['dry_run']:
            os.chdir(pkg["install_path"])

        if pkg['tools_path']:
            set_env(pkg)

        custom_run(pkg, "post_custom_run")
        if pkg['path']:
            custom_run(pkg, "post_custom_runl")

        pkg["logger"].info("Installation finished successfully!")

    if cfg["tools_sh_path"]:
        print('Installation finished, before invoking tools, source {}'.format(
            cfg["tools_sh_path"]))


def get_argparser():
    parser = argparse.ArgumentParser(prog="PyGears tools installer")

    parser.add_argument('-p',
                        dest='pkgs_fn',
                        default=os.path.join(os.path.dirname(__file__), 'pkgs.json'),
                        help="Path to packages description file")

    parser.add_argument('-o',
                        dest='tools_path',
                        default=None,
                        help="Directory to install tools to. Default will install system-wide")

    parser.add_argument('-w', dest='home_path', default=None, help="Directory to setup home in")

    parser.add_argument('-l',
                        dest='list_deps',
                        action='store_true',
                        help="Only list the package dependencies and exit")

    parser.add_argument('-d',
                        dest='install_deps',
                        action='store_true',
                        help="Automatically install system dependencies. Will require sudo.")

    parser.add_argument('-r', dest='dry_run', action='store_true', help="Dry run")

    parser.add_argument(
        'pkg_names',
        metavar='pkg_names',
        nargs='*',
        default=['gui'],
        help=
        'Names of packages to install. Can be one of: emacs, spacemacs, gtkwave, verilator, systemc, scv, pyenv, python, ripgrep, spacemacs_python, pygears'
    )

    return parser


def main(argv=sys.argv):
    parser = get_argparser()
    args = parser.parse_args(argv[1:])

    install(args.pkgs_fn, args.pkg_names, args.tools_path, args.home_path, args.install_deps,
            args.list_deps, args.dry_run)


# main([
#     '', '-d'
# ])
# main(['', '-dr'])
# main(['', '-dr'])



import configparser
import urllib.request
import tarfile
import io
import sys
import subprocess
import os
import shutil


def os_name():
    with open('/etc/os-release') as f:
        file_content = '[root]\n' + f.read()

    os_cfg = configparser.RawConfigParser()
    os_cfg.read_string(file_content)

    name = os_cfg['root']['NAME'].strip('"').split()[0].lower()

    return name


def get_url_archive_file_name(pkg):
    return os.path.basename(pkg["url"])


def get_url_archive_path(pkg):
    return os.path.join(pkg['install_path'], get_url_archive_file_name(pkg))


def download_source(pkg):
    fn = get_url_archive_file_name(pkg)

    def dlProgress(count, blockSize, totalSize):
        percent = int(count * blockSize * 100 / totalSize)
        sys.stdout.write("\rProgress: %d%%" % percent)
        sys.stdout.flush()

    pkg["logger"].info("Downloading " + pkg["url"])

    if not pkg['dry_run']:
        urllib.request.urlretrieve(pkg["url"], get_url_archive_path(pkg), dlProgress)
        sys.stdout.write("\n")

    return fn


class ProgressFileObject(io.FileIO):
    def __init__(self, path, pkg, *args, **kwargs):
        self._total_size = os.path.getsize(path)
        self._fn = os.path.basename(path)
        self._last_percent = -1
        self._pkg = pkg
        io.FileIO.__init__(self, path, *args, **kwargs)

    def read(self, size):
        percent = int(self.tell() * 100 / self._total_size)
        if percent != self._last_percent:
            self._last_percent = percent
            # sys.stdout.write("\r[" + pkg["name"] + "] Downloading " + self._fn + "... [%d%%]" % percent)
            sys.stdout.write("\rProgress: %d%%" % percent)
            sys.stdout.flush()

        return io.FileIO.read(self, size)


def untar(pkg):
    pkg["logger"].info("Unpacking " + get_url_archive_file_name(pkg))
    if not pkg['dry_run']:
        tar = tarfile.open(fileobj=ProgressFileObject(get_url_archive_path(pkg), pkg))
        tar.extractall(pkg['install_path'])
        tar.close()
        sys.stdout.write("\n")


def download_and_untar(pkg):
    fn = get_url_archive_path(pkg)
    if not os.path.exists(fn):
        download_source(pkg)
    else:
        pkg["logger"].info("Source archive already available and will be reused.")

    src_dir = pkg["src_root_path"]

    unpack = False
    if not os.path.exists(src_dir):
        unpack = True
    else:
        srctime = os.path.getmtime(fn)
        tartime = os.path.getmtime(src_dir)

        if tartime < srctime:
            unpack = True

            if not pkg['dry_run']:
                shutil.rmtree(src_dir)

            pkg["logger"].info("Source older than archive, updating...")

    if unpack:
        untar(pkg)
    else:
        pkg["logger"].info("Sources already present and up to date.")


def list_pkg_deps(pkgs):
    deps = []
    for pkg in pkgs:
        if 'git' in pkg:
            pkg['deps'] = ' '.join([pkg.get('deps', ''), 'git'])

        if 'deps' in pkg:
            deps += pkg['deps'].split()

    resp = [set(deps)]
    if any(pkg.get("flow", '') == "default_cpp" for pkg in pkgs):
        from pygears_tools import default_cpp
        resp.insert(0, [default_cpp.dependencies[os_name()]])

    return resp


def install_deps(pkgs):
    print("Installing dependencies")
    deps = list_pkg_deps(pkgs)

    for d in deps:
        cmd = "sudo apt install -y " + ' '.join(d)
        print(cmd)

        if not pkgs[0]['dry_run']:
            subprocess.run(cmd, shell=True, check=True)


def configure(pkg):
    if not pkg['dry_run']:
        os.chdir(pkg["src_root_path"])

    if "configure" in pkg:
        pkg["logger"].info("Running auto configure. Output redirected to configure.log .")
        if pkg['path']:
            cmd = f"./configure --prefix={pkg['path']} {pkg['configure']} > ../configure.log 2>&1"
        else:
            cmd = f"./configure {pkg['configure']} > ../configure.log 2>&1"

        pkg["logger"].info(f"Command: {cmd}.")

        if not pkg['dry_run']:
            subprocess.check_output(cmd, shell=True)

    if not pkg['dry_run']:
        os.chdir(pkg["install_path"])


def cmake(pkg):
    if not pkg['dry_run']:
        os.chdir(pkg["src_root_path"])
        if os.path.exists("build"):
            shutil.rmtree("build")

        os.mkdir("build")
        os.chdir("build")

    if "cmake_cfg" not in pkg:
        pkg["cmake_cfg"] = ""

    pkg["logger"].info("Running CMake. Output redirected to cmake.log .")

    if pkg['path']:
        cmd = (f"cmake -DCMAKE_INSTALL_PREFIX:PATH={pkg['path']} {pkg['cmake_cfg']}"
               " .. > ../../cmake.log 2>&1")
    else:
        cmd = (f"cmake {pkg['cmake_cfg']} .. > ../../cmake.log 2>&1")

    pkg["logger"].info(f"Command: {cmd}.")

    if not pkg['dry_run']:
        subprocess.check_output(cmd, shell=True)

    pkg["src_root_path"] = os.getcwd()

    if not pkg['dry_run']:
        os.chdir(pkg["src_root_path"])


def make(pkg):
    if not pkg['dry_run']:
        os.chdir(pkg["src_root_path"])

    pkg["logger"].info("Running make. Output redirected to make.log .")
    cmd = f"make -j4 > ../make.log 2>&1"
    pkg["logger"].info(f"Command: {cmd}.")

    if not pkg['dry_run']:
        try:
            subprocess.check_output(cmd, shell=True)
        except subprocess.CalledProcessError:
            pkg["logger"].error("Make finished with error, please check the log.")

    pkg["logger"].info("Running make install. Output redirected to make_install.log .")
    if pkg['path']:
        cmd = "make install > ../make_install.log 2>&1"
    else:
        cmd = "sudo make install > ../make_install.log 2>&1"

    pkg["logger"].info(f"Command: {cmd}.")

    if not pkg['dry_run']:
        try:
            subprocess.check_output(cmd, shell=True)
        except subprocess.CalledProcessError:
            pkg["logger"].error("Make install finished with error, please check the log.")

    if not pkg['dry_run']:
        os.chdir(pkg["src_root_path"])


def clone_git(pkg):
    clone_dir = pkg["src_root_path"]
    if not os.path.exists(clone_dir):
        pkg["logger"].info("Cloning git repo. Output redirected to git_clone.log .")
        cmd = f"git clone {pkg['git']} {clone_dir} > git_clone.log 2>&1"
        pkg["logger"].info(f"Command: {cmd}.")

        if not pkg['dry_run']:
            subprocess.check_output(cmd, shell=True)
    else:
        pkg["logger"].info("Source git repo already available and will be reused.")


def shell_source(script):
    """Sometime you want to emulate the action of "source" in bash,
    settings some environment variables. Here is a way to do it."""

    pipe = subprocess.Popen(". %s; env" % script,
                            stdout=subprocess.PIPE,
                            shell=True,
                            executable="/bin/bash" if os_name() == 'ubuntu' else None)
    output = pipe.communicate()[0].decode()
    env = {}
    for line in output.splitlines():
        try:
            keyval = line.split("=", 1)
            env[keyval[0]] = keyval[1]
        except:
            pass

    os.environ.update(env)


def set_env(pkg):
    if "env" not in pkg:
        return True

    pkg["logger"].info("Exporting the environment variables.")
    if not pkg['dry_run']:
        with open(pkg["tools_sh_path"], "a") as text_file:
            print("", file=text_file)
            print("# Environment for {}".format(pkg["name"]), file=text_file)
            for cmd in pkg["env"]:
                print(cmd.format(**pkg), file=text_file)

        shell_source(pkg["tools_sh_path"])


def custom_run(pkg, cmd_set):
    if cmd_set not in pkg:
        return True

    pkg["logger"].info("Running custom package commands. Output redirected to custom_cmd.log .")

    log_file = os.path.join(pkg["install_path"], "custom_cmd.log")
    for cmd in pkg[cmd_set]:
        cmd = cmd.format(**pkg)
        pkg["logger"].info('Command: "{}"'.format(cmd))

        if not pkg['dry_run']:
            subprocess.check_output("{} > {} 2>&1".format(cmd, log_file), shell=True)

def pip_install(pkg, pyenv):
    if pyenv:
        cmd = f'pip3 install -U {pkg["pip"]}'
    else:
        cmd = f'sudo pip3 install -U {pkg["pip"]}'

    log_file = os.path.join(pkg["install_path"], "pip.log")
    pkg["logger"].info('Command: "{}"'.format(cmd))
    if not pkg['dry_run']:
        subprocess.check_output(f"{cmd} > {log_file} 2>&1", shell=True)



from setuptools import setup, find_packages


setup(
    name='pygears-tools',
    version='0.1.14',
    description='Helper scripts for installing pygears related tools',

    # The project's main homepage.
    url='https://github.com/bogdanvuk/pygears-tools',

    # Author details
    author='Bogdan Vukobratovic',
    author_email='bogdan.vukobratovic@gmail.com',

    # Choose your license
    license='MIT',

    packages=find_packages(exclude=['examples*', 'docs']),
    package_data={'': ['*.json', '.spacemacs', 'gtkwave/*.diff']},
    include_package_data=True,
    entry_points={
        'console_scripts': [
            'pygears-tools-install = pygears_tools.install:main'
        ],
    }
)
